// SUPPLEMENTARY MATERIAL FOR ABREVAYA, HAUSMAN, AND KHAN ECONOMETRICA PAPER 
// "TESTING FOR CAUSAL EFFECTS IN A GENERALIZED REGRESSION MODEL WITH ENDOGENOUS REGRESSORS"
// OX CODE FOR MONTE CARLO SIMULATIONS (DESIGN 1, TABLE S1 IN PAPER)

#include <oxstd.h>
#import <maximize>

decl Rho1, Z1, Z2, Y1, Y2, Index, V;

// log-likelihood function for probit
fProbit1(const beta, const adFunc, const avScore, const amHessian)
{
    decl prob, weight;
	
    prob = probn((Z1~Z2)*beta);
    adFunc[0] = double(meanc(Y2.*log(prob)+(1-Y2).*log(1-prob)));

    if (avScore)
    {
        weight = (Y2-prob).*densn((Z1~Z2)*beta)./(prob.*(1-prob));
        avScore[0] = meanc(weight.*(Z1~Z2))';
    }
	return 1;                           // 1 indicates success
}

// log-likelihood function for probit 
fProbit2(const beta, const adFunc, const avScore, const amHessian)
{
	decl prob, weight;
	
    prob = probn((Z1~Y2)*beta);
    adFunc[0] = double(meanc(Y1.*log(prob)+(1-Y1).*log(1-prob)));

    if (avScore)
    {
        weight = (Y1-prob).*densn((Z1~Y2)*beta)./(prob.*(1-prob));
        avScore[0] = meanc(weight.*(Z1~Y2))';
    }
	return 1;                           // 1 indicates success
}

// auxiliary function for the full MLE
prob1_aux(const delta)
{
	decl temp;

	temp = (probn( ((Z1~Y2)*delta + Rho1*V) / ((1-Rho1^2)^0.5) )).*(V.>-Index);
	temp = meanr(temp)./probn(Index);
	return(temp);
}

// auxiliary function for the full MLE
prob2_aux(const delta)
{
	decl temp;

	temp = (probn( ((Z1~Y2)*delta + Rho1*V) / ((1-Rho1^2)^0.5) )).*(V.<-Index);
	temp = meanr(temp)./(1-probn(Index));
	return(temp);
}

// log-likelihood for probit with endogeneity
stage2_mle(delta, const adFunc, const avScore, const amHessian)
{
	decl prob1, prob2, temp1, temp2, temp;

	Rho1 = delta[0];
	delta = delta[1:];

	prob1 = prob1_aux(delta);
	prob2 = prob2_aux(delta);

	prob1 = (prob1.<=0).*(0.000001) + (prob1.>=1).*(0.999999) + (prob1.>0).*(prob1.<1).*prob1;
	prob2 = (prob2.<=0).*(0.000001) + (prob2.>=1).*(0.999999) + (prob2.>0).*(prob2.<1).*prob2;

    adFunc[0] = double(meanc(Y1.*Y2.*log(prob1)+(1-Y1).*Y2.*log(1-prob1)+ Y1.*(1-Y2).*log(prob2)+(1-Y1).*(1-Y2).*log(1-prob2)));

	return 1;
}

// sign function: 1 if positive, 0 otherwise
sign(v)
{
	return( (v.>0)-(v.<0) );
}

// Kendall's tau (rank correlation statistic) --- used for test with no additional covariates
tau_function()
{
	decl i, j, numer, denom, nobs, tau_stat, zeta_val, temp;

	numer = 0;
	denom = 0;
	nobs = rows(Y1);
	
	for (i=0; i<nobs-1; i++)
		for (j=i+1; j<nobs; j++)
		{
			denom ++;
			numer += sign(Y1[i]-Y1[j])*sign(Z2[i]-Z2[j]);
		}

	tau_stat = numer/denom;

	zeta_val = 0;
	for (i=0; i<nobs-1; i++)
	{
		temp = sign(Y1[i]-Y1)'*sign(Z2[i]-Z2);
		zeta_val += ((temp/(nobs-1))-tau_stat)^2;
	}
	zeta_val = zeta_val/nobs;

	return(tau_stat|(2*sqrt(zeta_val/nobs)));
}

// function to make data --- see paper for design details
makedata(const nobs, const rho, const beta_value)
{
	decl u1, v2;
	
	Z1=ones(nobs,1);

	Z2 = ((rann(nobs,1).^2)+(rann(nobs,1).^2)+(rann(nobs,1).^2))/sqrt(6);   // normalized chi-square 3 d.f.
	Z2 = Z2-(3/sqrt(6));   // get mean zero
	
	v2 = rann(nobs,1);
	u1 = rho*v2 + sqrt(1-rho^2)*rann(nobs,1);

	Y2 = (Z2 + v2 .> 0);
	Y1 = (beta_value*Y2 + u1 .> 0);
	
	return(1);
}

main()
{
	decl ir, delta2, dfunc, mhess, cov_delta2, nobs;
	decl delta1, cov_delta1;
	decl beta_iv, cov_iv, eps_iv, sigma2_iv;
	decl temp_tau;
	decl outvec = 0;
	decl num_simul = 0;
	decl simul_type = 0;
	decl rho;
	decl beta_value;  // slope on the "endogenous" indicator variable
	decl total_simuls = 1000;
	decl total_simul_types = 16;
	decl i, temp_matrix, xhat;
	
	nobs = 500;
	
	for (simul_type = 1; simul_type<=total_simul_types; simul_type++)
	{
		println("Simulation design ", simul_type);
		
		outvec=0;
		if (simul_type == 1)
		{
			nobs = 500;
			rho = 0;
			beta_value = 0;
		}
		else if (simul_type == 2)
		{
			nobs = 500;
			rho = 0.25;
			beta_value = 0;
		}
		else if (simul_type == 3)
		{
			nobs = 500;
			rho = 0.50;
			beta_value = 0;
		}
		else if (simul_type == 4)
		{
			nobs = 500;
			rho = 0.75;
			beta_value = 0;
		}
		else if (simul_type == 5)
		{
			nobs = 500;
			rho = 0;
			beta_value = 0.1;
		}
		else if (simul_type == 6)
		{
			nobs = 500;
			rho = 0.25;
			beta_value = 0.1;
		}
		else if (simul_type == 7)
		{
			nobs = 500;
			rho = 0.50;
			beta_value = 0.1;
		}
		else if (simul_type == 8)
		{
			nobs = 500;
			rho = 0.75;
			beta_value = 0.1;
		}
		else if (simul_type == 9)
		{
			nobs = 500;
			rho = 0;
			beta_value = 0.2;
		}
		else if (simul_type == 10)
		{
			nobs = 500;
			rho = 0.25;
			beta_value = 0.2;
		}
		else if (simul_type == 11)
		{
			nobs = 500;
			rho = 0.50;
			beta_value = 0.2;
		}
		else if (simul_type == 12)
		{
			nobs = 500;
			rho = 0.75;
			beta_value = 0.2;
		}
		else if (simul_type == 13)
		{
			nobs = 500;
			rho = 0;
			beta_value = 0.3;
		}
		else if (simul_type == 14)
		{
			nobs = 500;
			rho = 0.25;
			beta_value = 0.3;
		}
		else if (simul_type == 15)
		{
			nobs = 500;
			rho = 0.50;
			beta_value = 0.3;
		}
		else if (simul_type == 16)
		{
			nobs = 500;
			rho = 0.75;
			beta_value = 0.3;
		}

		for (num_simul = 1; num_simul<=total_simuls; num_simul++)
		{
			println("Simulation number: ", num_simul);
			println("Number of observations: ", nobs);
			println("rho value: ", rho);
		
			makedata(nobs,rho,beta_value);

			delta2 = invert((Z1~Z2)'(Z1~Z2))*((Z1~Z2)'Y2);

			ir = MaxBFGS(fProbit1, &delta2, &dfunc, 0, FALSE);
			if (ir == MAX_CONV || ir == MAX_WEAK_CONV)
			    if (Num2Derivative(fProbit1, delta2, &mhess))
					cov_delta2 = -invert(mhess) / nobs;

			print(delta2~sqrt(diagonal(cov_delta2)'));

			Index = (Z1~Z2)*delta2;
			delta1 = invert((Z1~Y2)'(Z1~Y2))*((Z1~Y2)'Y1);
			ir = MaxBFGS(fProbit2, &delta1, &dfunc, 0, FALSE);

			V = quann(range(0.01,0.99,0.005)')';
			Rho1 = rho;
			delta1 = Rho1|delta1;

			ir = MaxBFGS(stage2_mle, &delta1, &dfunc, 0, TRUE);
			if (ir == MAX_CONV || ir == MAX_WEAK_CONV)
			    if (Num2Derivative(stage2_mle, delta1, &mhess))
					cov_delta1 = -invert(mhess) / nobs;
			if (cov_delta1==0)				// if decomposition fails, var matrix equal to zeros
				cov_delta1=zeros(3,3);

			print("MLE", delta1~sqrt(diagonal(cov_delta1)'));

			// IV estimation
			beta_iv = invert((Z1~Z2)'(Z1~Y2))*((Z1~Z2)'Y1);

			eps_iv = Y1-(Z1~Y2)*beta_iv;
			xhat = (Z1~Z2)*(invert((Z1~Z2)'(Z1~Z2))*(Z1~Z2)'(Z1~Y2));
			temp_matrix = zeros(columns(xhat),columns(xhat));
			for (i=0; i<nobs; i++)
				temp_matrix += eps_iv[i]*eps_iv[i]*xhat[i][]'xhat[i][];
			cov_iv = invert(xhat'xhat)*temp_matrix*invert(xhat'xhat);
		
			print("IV", beta_iv[rows(beta_iv)-1]~sqrt(cov_iv[rows(beta_iv)-1][rows(beta_iv)-1]));
		
			// semiparametric (kendall's tau) estimation
			temp_tau = tau_function();
			print("TAU", temp_tau);

			if (outvec==0)
				outvec = delta1[2]|sqrt(cov_delta1[2][2])|beta_iv[1]|sqrt(cov_iv[1][1])|temp_tau;
			else
				outvec = outvec~(delta1[2]|sqrt(cov_delta1[2][2])|beta_iv[1]|sqrt(cov_iv[1][1])|temp_tau);
		
			if (simul_type == 1)
				savemat("outvec-monte-carlo-simul1.mat", outvec);
			else if (simul_type == 2)
				savemat("outvec-monte-carlo-simul2.mat", outvec);
			else if (simul_type == 3)
				savemat("outvec-monte-carlo-simul3.mat", outvec);
			else if (simul_type == 4)
				savemat("outvec-monte-carlo-simul4.mat", outvec);
			else if (simul_type == 5)
				savemat("outvec-monte-carlo-simul5.mat", outvec);
			else if (simul_type == 6)
				savemat("outvec-monte-carlo-simul6.mat", outvec);
			else if (simul_type == 7)
				savemat("outvec-monte-carlo-simul7.mat", outvec);
			else if (simul_type == 8)
				savemat("outvec-monte-carlo-simul8.mat", outvec);
			else if (simul_type == 9)
				savemat("outvec-monte-carlo-simul9.mat", outvec);
			else if (simul_type == 10)
				savemat("outvec-monte-carlo-simul10.mat", outvec);
			else if (simul_type == 11)
				savemat("outvec-monte-carlo-simul11.mat", outvec);
			else if (simul_type == 12)
				savemat("outvec-monte-carlo-simul12.mat", outvec);
			else if (simul_type == 13)
				savemat("outvec-monte-carlo-simul13.mat", outvec);
			else if (simul_type == 14)
				savemat("outvec-monte-carlo-simul14.mat", outvec);
			else if (simul_type == 15)
				savemat("outvec-monte-carlo-simul15.mat", outvec);
			else if (simul_type == 16)
				savemat("outvec-monte-carlo-simul16.mat", outvec);

		}
	}
}
