// SUPPLEMENTARY MATERIAL FOR ABREVAYA, HAUSMAN, AND KHAN ECONOMETRICA PAPER 
// "TESTING FOR CAUSAL EFFECTS IN A GENERALIZED REGRESSION MODEL WITH ENDOGENOUS REGRESSORS"
// OX CODE FOR MONTE CARLO SIMULATIONS (DESIGN 2, TABLE S2 IN PAPER)

#include <oxstd.h>
#import <maximize>

decl Z1, Z2, Y1, Y2, Index, h;

// log-likelihood function for probit
fProbit1(const beta, const adFunc, const avScore, const amHessian)
{
    decl prob, weight;
	
    prob = probn((Z1~Z2)*beta);
	
    adFunc[0] = meanc(Y2.*log(prob+0.00001)+(1-Y2).*log(1-prob+0.00001));

    if (avScore)
    {
        weight = (Y2-prob).*densn((Z1~Z2)*beta)./(prob.*(1-prob));
        avScore[0] = meanc(weight.*(Z1~Z2))';
    }

	return 1;                           // 1 indicates success
}

// sign function: 1 if positive, 0 otherwise
sign(v)
{
	return( (v.>0)-(v.<0) );
}

// Kendall's tau (rank correlation statistic) --- used for test with no additional covariates
tau_function()
{
	decl i, j, numer, denom, nobs, tau_stat, zeta_val, temp;

	numer = 0;
	denom = 0;
	nobs = rows(Y1);
	
	for (i=0; i<nobs-1; i++)
		for (j=i+1; j<nobs; j++)
		{
			denom ++;
			numer += sign(Y1[i]-Y1[j])*sign(Z2[i]-Z2[j]);
		}

	tau_stat = numer/denom;

	zeta_val = 0;
	for (i=0; i<nobs-1; i++)
	{
		temp = sign(Y1[i]-Y1)'*sign(Z2[i]-Z2);
		zeta_val += ((temp/(nobs-1))-tau_stat)^2;
	}
	zeta_val = zeta_val/nobs;

	return(tau_stat|(2*sqrt(zeta_val/nobs)));
}

// rank test statistic in the presence of an additional covariate
tau_function_z1()
{
	decl i, j, numer, denom, nobs, tau_stat, zeta_val, temp;

	numer = 0;
	denom = 0;
	nobs = rows(Y1);
	
	for (i=0; i<nobs-1; i++)
		for (j=i+1; j<nobs; j++)
		{
			denom += (densn( (Z1[i][1]-Z1[j][1]) / h) );
			numer += (densn( (Z1[i][1]-Z1[j][1]) / h)*sign(Y1[i]-Y1[j])*sign(Index[i]-Index[j]));
		}

	tau_stat = numer/denom;

	decl Y1temp, Y2temp, Z1temp, Z2temp, tempidx, bsiters = 50, delta2, cov_delta2, ir, k, dfunc, mhess, bs_stats;

	bs_stats = zeros(bsiters,1);
	
	for (k=0; k<bsiters; k++)
	{
		tempidx = round(ranu(rows(Y1),1)*rows(Y1)-0.5);
		Y1temp = Y1;
		Y2temp = Y2;
		Z1temp = Z1;
		Z2temp = Z2;
		
		Y1 = Y1[tempidx][];
		Y2 = Y2[tempidx][];
		Z1 = Z1[tempidx][];
		Z2 = Z2[tempidx][];

		delta2 = -5|1|1;
		ir = MaxBFGS(fProbit1, &delta2, &dfunc, 0, TRUE);
		if (ir == MAX_CONV || ir == MAX_WEAK_CONV)
		    if (Num2Derivative(fProbit1, delta2, &mhess))
				cov_delta2 = -invert(mhess) / nobs;

		Index = (Z1~Z2)*delta2;
		numer = 0;
		denom = 0;
		nobs = rows(Y1);
		for (i=0; i<nobs-1; i++)
			for (j=i+1; j<nobs; j++)
			{
				denom += (densn( (Z1[i][1]-Z1[j][1]) / h) );
				numer += (densn( (Z1[i][1]-Z1[j][1]) / h)*sign(Y1[i]-Y1[j])*sign(Index[i]-Index[j]));
			}

		bs_stats[k] = numer/denom;
		Y1 = Y1temp;
		Y2 = Y2temp;
		Z1 = Z1temp;
		Z2 = Z2temp;
	}
	
	return(tau_stat|sqrt(variance(bs_stats)));
}


// function to make data --- see paper for design details
makedata(const nobs, const rho, const beta_value)
{
	decl u1, v2;
	
	Z1 = 1~((rann(nobs,1).^2)+(rann(nobs,1).^2)+(rann(nobs,1).^2));
	Z2 = 4*ranu(nobs,1);
	Y2 = (Z1[][1] + Z2[][] - 5 +rann(nobs,1).>0);
	Y1 = (Z1[][1] - 3 + rann(nobs,1) .> 0);

	return(1);
}

main()
{
	decl ir, delta2, dfunc, mhess, cov_delta2, nobs;
	decl delta1, cov_delta1;
	decl beta_iv, cov_iv, eps_iv, sigma2_iv;
	decl temp_tau;
	decl outvec = 0;
	
	decl num_simul = 0;
	decl simul_type = 0;
	decl rho;
	decl beta_value;
	decl total_simuls = 1000;
	decl total_simul_types = 16;

	decl i, temp_matrix, xhat;
	
	rho = 0.5;
	beta_value = 0.5;
	
	decl beta_ols, eps_ols, cov_ols;

	decl simul_ctr;

	for (simul_ctr=1; simul_ctr<10; simul_ctr++)
	{
		if (simul_ctr==1)
		{
			nobs = 50;
			h = 0.1;
		}
		else if (simul_ctr==2)
		{
			nobs = 50;
			h=0.3;
		}
		else if (simul_ctr==3)
		{
			nobs = 100;
			h = 0.1;
		}
		else if (simul_ctr==4)
		{
			nobs = 100;
			h=0.3;
		}
		else if (simul_ctr==5)
		{
			nobs = 200;
			h = 0.1;
		}
		else if (simul_ctr==6)
		{
			nobs = 200;
			h=0.3;
		}
		else if (simul_ctr==7)
		{
			nobs = 400;
			h = 0.1;
		}
		else if (simul_ctr==8)
		{
			nobs = 400;
			h=0.3;
		}
		else if (simul_ctr==9)
		{
			nobs = 1000;
			h = 0.1;
		}
		else if (simul_ctr==10)
		{
			nobs = 1000;
			h=0.3;
		}
		
		println("Simulation for n = ", nobs, " bandwidth = ", h);
		outvec = 0;

		for (num_simul = 1; num_simul<=total_simuls; num_simul++)
		{
			println("Simulation number: ", num_simul);
			println("Number of observations: ", nobs);
			makedata(nobs,rho,beta_value);
		
			// OLS estimation
			beta_ols = invert((Z1~Y2)'(Z1~Y2))*((Z1~Y2)'Y1);
			eps_ols = Y1-(Z1~Y2)*beta_ols;

			temp_matrix = zeros(columns((Z1~Y2)),columns((Z1~Y2)));
			for (i=0; i<nobs; i++)
				temp_matrix += eps_ols[i]*eps_ols[i]*(Z1[i][]~Y2[i][])'(Z1[i][]~Y2[i][]);
			cov_ols = invert((Z1~Y2)'(Z1~Y2))*temp_matrix*invert((Z1~Y2)'(Z1~Y2));
		
			print("OLS", beta_ols[rows(beta_ols)-1]~sqrt(cov_ols[rows(beta_ols)-1][rows(beta_ols)-1]));

			// IV estimation
			beta_iv = invert((Z1~Z2)'(Z1~Y2))*((Z1~Z2)'Y1);
			eps_iv = Y1-(Z1~Y2)*beta_iv;

			temp_matrix = zeros(columns((Z1~Y2)),columns((Z1~Y2)));
			xhat = (Z1~Z2)*invert((Z1~Z2)'(Z1~Y2))*((Z1~Z2)'(Z1~Y2));
			for (i=0; i<nobs; i++)
				temp_matrix += eps_iv[i]*eps_iv[i]*(Z1[i][]~Y2[i][])'(Z1[i][]~Y2[i][]);
			cov_iv = invert((Z1~Y2)'(Z1~Y2))*temp_matrix*invert((Z1~Y2)'(Z1~Y2));
		
			print("IV", beta_iv[rows(beta_iv)-1]~sqrt(cov_iv[rows(beta_iv)-1][rows(beta_iv)-1]));
		
			// semiparametric (kendall's tau) estimation

			// probit estimation first stage
			delta2 = -5|1|1;

			ir = MaxBFGS(fProbit1, &delta2, &dfunc, 0, TRUE);
			if (ir == MAX_CONV || ir == MAX_WEAK_CONV)
			    if (Num2Derivative(fProbit1, delta2, &mhess))
					cov_delta2 = -invert(mhess) / nobs;
			print(delta2~sqrt(diagonal(cov_delta2)'));

			Index = (Z1~Z2)*delta2;
			temp_tau = tau_function_z1();
			print("TAU", temp_tau);

			if (outvec==0)
				outvec = beta_ols[rows(beta_ols)-1]|sqrt(cov_ols[rows(beta_ols)-1][rows(beta_ols)-1])|beta_iv[rows(beta_iv)-1]|sqrt(cov_iv[rows(beta_iv)-1][rows(beta_iv)-1])|temp_tau;
			else
				outvec = outvec~(beta_ols[rows(beta_ols)-1]|sqrt(cov_ols[rows(beta_ols)-1][rows(beta_ols)-1])|beta_iv[rows(beta_iv)-1]|sqrt(cov_iv[rows(beta_iv)-1][rows(beta_iv)-1])|temp_tau);
		}

		println(meanr(outvec));
		println(sqrt(varr(outvec)));

		println("rejections at 5% level", sumr( fabs(outvec[0][]./outvec[1][]).>1.96 ), " ", sumr( fabs(outvec[2][]./outvec[3][]).>1.96 ), " ", sumr( fabs(outvec[4][]./outvec[5][]).>1.96 ));
		println("rejections at 10% level", sumr( fabs(outvec[0][]./outvec[1][]).>1.645 ), " ", sumr( fabs(outvec[2][]./outvec[3][]).>1.645 ), " ", sumr( fabs(outvec[4][]./outvec[5][]).>1.645 ));
	}
}
