// SUPPLEMENTARY MATERIAL FOR ABREVAYA, HAUSMAN, AND KHAN ECONOMETRICA PAPER 
// "TESTING FOR CAUSAL EFFECTS IN A GENERALIZED REGRESSION MODEL WITH ENDOGENOUS REGRESSORS"
// OX CODE FOR EMPIRICAL APPLICATION (TABLE S7 IN PAPER)
//
// NOTE THAT THIS FILE DOES THE SEMIPARAMETRIC ESTIMATION FOR THE FULL SET OF COVARIATES
// (EDUC, MOTHER'S AGE, AGE OF 1ST CHILD --- LAST ROW IN TABLE S7); THE OTHER RESULTS CAN
// BE OBTAINED BY CHANGING THE SET OF COVARIATES IN THE MAIN() FUNCTION

#include <oxstd.h>
#include <oxprob.h>
#import <maximize>

decl Y, X, prY, prX;
decl y, p, psum, ndim, ftol, lam1, lam2, nfunk;
decl S2M, stage2sort = 0;
decl S3M;
decl h_S2, h_S3;   // window widths for stage 2 and 3, respectively
decl num_NN;

// log-likelihood function for probit
fProbit(const beta, const adFunc, const avScore, const amHessian)
{
	decl prob, weight;
	
    prob = probn(prX*beta);

    adFunc[0] = double(meanc(prY.*log(prob)+(1-prY).*log(1-prob)));

    if (avScore)
    {
        weight = (prY-prob).*densn(prX*beta)./(prob.*(1-prob));
        avScore[0] = meanc(weight.*prX)';
    }

	return 1;                           // 1 indicates success
}

// original (kernel-based weight) version of the stage 2 objective function;
// note that there are inserted lines that speed things up for the specific
// case of a binary choice model (can be removed for other models)
//
// global variable S2M ("stage 2 matrix") has the following columns:
// column 0: D1
// column 1: D2
// column 2: stage 1 estimated index (w'deltahat)
// column 3+: X variables
//
// global flag stage2sort indicates whether D1's have been sorted
stage2(theta)
{
	decl cutoff = rows(S2M)-sumc(S2M[][0]);
	decl temp0, temp1, i, j, retval, nobs;

	decl time_stamp;

	time_stamp = timer();
	
	if (stage2sort==0)
		S2M = sortbyc(S2M,0);

	temp0 = sortbyc(S2M[0:cutoff-1][1:2]~(S2M[0:cutoff-1][3:]*theta),2);
	temp1 = sortbyc(S2M[cutoff:][1:2]~(S2M[cutoff:][3:]*theta),2);	

	retval = 0;
	
	// loop through D1=0 observations
	nobs = rows(temp0);
	for (i=0; i<nobs-1; i++)
	if (temp0[i][0]==0)   // this line speeds things up for binary choice model
	{
		j = i+1;

		// skip over index values that tie the i-th observation
		while ((temp0[i][2]==temp0[j][2]) && j<nobs-1)
			j++;

		// for larger index values, add to the objective function if D2's are ordered
		retval += (temp0[j:nobs-1][0].>temp0[i][0])'*densn((temp0[j:nobs-1][1]-temp0[i][1])/h_S2);
	}	

	// loop through D1=1 observations
	nobs = rows(temp1);
	for (i=0; i<nobs-1; i++)
	if (temp1[i][0]==0)   // this line speeds things up for binary choice model
	{
		j = i+1;

		// skip over index values that tie the i-th observation
		while ((temp1[i][2]==temp1[j][2]) && j<nobs-1)
			j++;

		// for larger index values, add to the objective function if D2's are ordered
		retval += (temp1[j:nobs-1][0].>temp1[i][0])'*densn((temp1[j:nobs-1][1]-temp1[i][1])/h_S2);
	}

	return(-retval);
}

// An alternative version of the stage 2 objective function that uses a nearest neighbor
// approach in order to cut down on the computation time
stage2_nn(theta)
{
	decl cutoff = rows(S2M)-sumc(S2M[][0]);
	decl temp0, temp1, i, j, retval, nobs;

	decl time_stamp;

	time_stamp = timer();
	
	if (stage2sort==0)
		S2M = sortbyc(S2M,0);

	temp0 = sortbyc(S2M[0:cutoff-1][1:2]~(S2M[0:cutoff-1][3:]*theta),2);
	temp1 = sortbyc(S2M[cutoff:][1:2]~(S2M[cutoff:][3:]*theta),2);	

	retval = 0;
	
	// loop through D1=0 observations
	nobs = rows(temp0);
	for (i=0; i<nobs-1; i++)
	if (temp0[i][0]==0)   // this line speeds things up for binary choice model
	{
		j = i+num_NN;
		if (j>=nobs)
			j = nobs-1;

		// for larger index values, add to the objective function if D2's are ordered
		retval += sumc(temp0[i+1:j][0]);
	}	

	// loop through D1=1 observations
	nobs = rows(temp1);
	for (i=0; i<nobs-1; i++)
	if (temp1[i][0]==0)   // this line speeds things up for binary choice model
	{
		j = i+num_NN;
		if (j>=nobs)
			j = nobs-1;

		// for larger index values, add to the objective function if D2's are ordered
		retval += sumc(temp1[i+1:j][0]);
	}

	return(-retval);
}

// sign function: 1 if positive, 0 otherwise
sign(v)
{
	return( (v.>0)-(v.<0) );
}

// stage 3 rank correlation statistic
//
// global variable S3M ("stage 3 matrix") has the following columns:
// column 0: D2
// column 1: stage 1 estimated index (w'deltahat)
// column 2: stage 2 estimated index (x'betahat)
stage3()
{
	decl i, j, numer, denom, nobs, tau_stat;

	numer = 0;
	denom = 0;
	nobs = rows(S3M);
	
	for (i=0; i<nobs-1; i++)
	{
		denom += sumc( densn((S3M[i][2]-S3M[i+1:nobs-1][2])/h_S3) );
		numer += sumc( densn((S3M[i][2]-S3M[i+1:nobs-1][2])/h_S3).*sign(S3M[i][0]-S3M[i+1:nobs-1][0]).*sign(S3M[i][1]-S3M[i+1:nobs-1][1]) );
	}
		
	tau_stat = numer/denom;
	return(tau_stat);
}

// NORMVEC:  normalize vector
normvec(const v)
{
	return(v./sqrt(v'v));
}

// AMOTRY:  auxiliary function for "amoeba"
amotry(func, const ihi, const fac)
{
	decl fac1, fac2, ytry, ptry;

	fac1 = (1-fac)/ndim;
	fac2 = fac1-fac;
	ptry = psum*fac1-p[ihi][]'fac2;
	ptry = normvec(ptry);
	ytry = func(ptry);

	if (ytry<y[ihi])
	{
		y[ihi] = ytry;
		p[ihi][] = ptry';
		psum = sumc(p)';
		psum = normvec(psum);
	}

	return(ytry);   
}

// AMOEBA:  auxiliary function for "nelder"
amoeba(func)
{
	decl temp, ihi, inhi, ilo, ysave, i, rtol, ytry;
	
	psum=sumc(p)';
	psum = normvec(psum);

	nfunk=0;
	
	while (1)  // endless loop until "break" encountered
	{
		temp = sortbyc(y~(range(0,ndim,1))',0);
		ihi = temp[rows(temp)-1][1];
		inhi = temp[rows(temp)-2][1];
		ilo = temp[0][1];

	    rtol=2*fabs(y[ihi]-y[ilo])/(fabs(y[ihi])+fabs(y[ilo]));

		if(rtol<ftol)
		{
			temp = y[0];
			y[0] = y[ilo];
			y[ilo] = temp;
			temp = p[0][];
			p[0][] = p[ilo][];
			p[ilo][] = temp;
			break;
		}

	//	if(nfunk>=5000)
	//	  print("Error");

		nfunk = nfunk+2;
		ytry = amotry(func,ihi,-1);

		if (ytry<y[ilo])
			ytry = amotry(func,ihi,2);
		else if (ytry>=y[inhi])
		{
			ysave = y[ihi];
			ytry = amotry(func,ihi,0.5);
			if (ytry>=ysave)
			{
				for (i=0; i<=ndim; i++)
					if (i!=ilo)
				    {
						p[i][] = normvec(0.5*(p[i][]'+p[ilo][]'))';
						y[i] = func(p[i][]');
					}
				psum = sumc(p)';
				psum = normvec(psum);
				nfunk = nfunk+ndim;
			}
		}
		else
			nfunk=nfunk-1;
			print(nfunk," ");
	}
	return(nfunk);	   
}


// Nelder-Mead function to minimize "func"
nelder(func, const start, const offset, const iters)
{
	decl i, j, oldp;

	ftol = 1e-06;
	lam1 = 0.6;
	lam2 = 0.4;

	p = zeros(ndim+1,ndim);
	y = zeros(ndim+1,1);
	
	p[0][] = normvec(start)';
	y[0] = func(p[0][]');

	for (i=1; i<=ndim; i++)
    {
		p[i][] = normvec(start+offset[i-1][]')';
		y[i] = func(p[i][]');
	}

	println("Starting value ", y[0]~p[0][]);
	println("Starting values ", y~p);

	for (j=1; j<=iters; j++)
	{
		oldp = p;
		amoeba(func);
		for (i=0;i<=ndim;i++)
		{
			p[i][] = normvec(p[0][]'+lam2*(oldp[i][]-oldp[0][])')';
			y[i] = func(p[i][]');
		}

		println("Iteration ", j);
		println("Obj func calls ", nfunk);
		println("Obj fn value ", -y[0]);
		println("Estimates ", p);

		ftol = ftol*lam1;
	}

	return(p[0][]');	
}

// function to call Nelder-Mead algorithm for the stage 2 objective function
stage2_nelder(start)
{
	ndim = columns(S2M)-3;
	return(nelder(stage2,start,start'.*ranu(rows(start),rows(start)),2));
}

// function to call Nelder-Mead algorithm for the stage 2 objective function (nearest neighbor version)
stage2_nelder_nn(start)
{
	ndim = columns(S2M)-3;
	return(nelder(stage2_nn,start,start'.*ranu(rows(start),rows(start)),3));
}

main()
{
	decl gl_d = loadmat("ae2000.txt");
	decl d;
	decl nobs;
	decl temp, temp2, temp3;
    decl b_probit, ir, mhess, dfunc, cov_probit;
	decl bs_size = 20000;
	decl outvec = 0;
	decl iteration_number = 1;
	decl i, j, max_theta, temp_theta, temp_stage2, max_stage2, temp_mult;
	
	for (iteration_number=1; iteration_number<=200; iteration_number++)
	{
		println("Bootstrap iteration #", iteration_number);
	
		d = gl_d[round(ranu(bs_size,1)*rows(gl_d)-0.5)][];
	//	d = gl_d;       // use d=gl_d in order to get estimates for the original sample

		nobs = rows(d);
		prY = d[][2];   			// first-stage Y variable = hadthird
		prX = 1~d[][3:6]~d[][9];	// first-stage X vars = samesex, ageatfirstbirth, educ, age of 1st child
	
		b_probit = zeros(columns(prX),1);        // starting values
	    ir = MaxBFGS(fProbit, &b_probit, &dfunc, 0, FALSE);
	
		if (ir == MAX_CONV || ir == MAX_WEAK_CONV)
	        if (Num2Derivative(fProbit, b_probit, &mhess))
				cov_probit = -invert(mhess) / nobs;
	
		print("PROBIT OUTPUT\n");
		print("%r",{"const","samesex","age_firstbirth","child1_age","child2_age","educ"},
		      "%c",{"Coeff","s.e."},
			  "%#10.3g",b_probit~sqrt(diagonal(cov_probit)'));


		// global variable S2M ("stage 2 matrix") has the following columns:
		// column 0: D1
		// column 1: D2
		// column 2: stage 1 estimated index (w'deltahat)
		// column 3+: X variables
		S2M = d[][2]~d[][0]~(prX*b_probit)~prX[][2:];

		temp2 = invert((1~d[][2]~prX[][2:])'(1~d[][2]~prX[][2:]))*((1~d[][2]~prX[][2:])'d[][0]);
		temp2 = normvec(temp2[2:]);

		h_S2 = sqrt(varc(S2M[][2]))*(nobs^(-1/5));
		println("Bandwidth for stage two: ", h_S2);

		//num_NN = round((h_S2/(quantilec(S2M[][2],0.75)-quantilec(S2M[][2],0.25)))*(rows(S2M)/2));
		//
		//println("Number of nearest neighbors for stage two: ", num_NN);

		// find an initial starting value
		max_theta = temp2;
		max_stage2 = -stage2(temp2);
		println("Initial theta: ", temp2');
		println("Initial obj. func. value: ", max_stage2);

		temp2 = normvec(-0.0065|0.0137|0.0005|0.0313);
		temp_stage2 = -stage2(temp2);
		if (temp_stage2 > max_stage2)
		{
			max_theta = temp_theta;
			max_stage2 = temp_stage2;
		}

		for (i=0; i<100; i++)
		{
			temp_theta = normvec(temp2.*(0.7+(0.3*ranu(rows(temp2),1))));
			temp_stage2 = -stage2(temp_theta);
			if (temp_stage2 > max_stage2)
			{
				max_theta = temp_theta;
				max_stage2 = temp_stage2;
			}
		}
		temp2 = max_theta;
		print("Done with first initialization phase ...  obj. func. value: ", max_stage2);
		println("Last theta value: ", temp2');

		for (j=0; j<rows(temp2); j++)
			for (i=0; i<50; i++)
			{
				temp_mult = ones(rows(temp2),1);
				temp_mult[j] = -1+2*ranu(1,1);
				temp_theta = normvec(temp2.*temp_mult);
				temp_stage2 = -stage2(temp_theta);
				if (temp_stage2 > max_stage2)
				{
					max_theta = temp_theta;
					max_stage2 = temp_stage2;
				}
			}
		temp2 = max_theta;
		print("Done with second initialization phase ...  obj. func. value: ", max_stage2);
		println("Last theta value: ", temp2');
		
		temp2 = stage2_nelder(temp2);
		temp2 = stage2_nelder(temp2);
		temp2 = stage2_nelder(temp2);

		S3M = d[][0]~(prX*b_probit)~(prX[][2:]*temp2);
		h_S3 = sqrt(varc(S3M[][2]))*(rows(gl_d)^(-1/5));
	
		println("Bandwidth for stage three: ", h_S3);
		temp3 = stage3();
		println("Stage 3 estimate: ", temp3);
	
		if (outvec==0)
			outvec = (b_probit|temp2|temp3);
		else
			outvec = outvec~(b_probit|temp2|temp3);

		savemat("outvec.mat", outvec);
	}
}
